/*******************************************************************************
 * Program..: bedroom.ino HVAC thermostat for Bedroom
 * Author...: Jan-marie Newton
 * Date.....: 2019-08-22
 * History..: 
 * Notes....: version 5.0
 * LCD Layout: 16x2   char(223) degree  char(219) open box char(255) filled box
 *             Heat OOOO 100.0'   wall temperature
 *               SP:72'  100.0'   room temperature
 *             0123456789012345
 *******************************************************************************/
//include <avr/wdt.h>
#include <OneWire.h>
#include <DallasTemperature.h>
#include <SPI.h>         // needed for Arduino versions later than 0018
#include <Ethernet.h>
#include <EthernetUdp.h> // UDP library from: bjoern@cs.stanford.edu 12/30/2008
#include <TimedAction.h>
#include <Encoder.h>     // for rotary nob
#include <LiquidCrystal_I2C.h> // https://create.arduino.cc/projecthub/akshayjoseph666/interface-i2c-16x2-lcd-with-arduino-uno-just-4-wires-273b24
#define ENCODER_DO_NOT_USE_INTERRUPTS
#define SP_STEP 8

// Constants - set relay pin numbers:
//------------------------------------------------------------------------------
#define SSL_RELAY 4
#define ONE_WIRE_BUS 5 // digital pin we're connected to

// Declaration for I2C 1602 LCD Display Module 16X2 Character Serial Blue
// on the MEGA use pin 20 (SDA) and 21 (SCL).
//------------------------------------------------------------------------------
LiquidCrystal_I2C lcd(0x27, 16, 2);

// 360 Degree Rotary Encoder Code Switch Push Button EC11 Digital Potentiometer
// Best Performance: both pins have interrupt capability
//------------------------------------------------------------------------------
Encoder spRotary(3, 2);

// Setup a oneWire instance to communicate with any OneWire devices  
//------------------------------------------------------------------------------
// (not just Maxim/Dallas temperature ICs) 
OneWire oneWire(ONE_WIRE_BUS);       //  temp sensor DS18B20
DallasTemperature sensors(&oneWire); // Pass our oneWire reference to Dallas Temperature


// Initialize Ethernet card
//------------------------------------------------------------------------------
EthernetUDP Udp;         // An EthernetUDP instance to let us send and receive packets over UDP
IPAddress ip(192, 168, 1, 124);
byte mac[]             = {0xDE, 0xAD, 0xBE, 0xEF, 0x1E, 0x1B}; // Enter a MAC address and IP address
unsigned int localPort = 8888;      // local port to listen on

// constants won't change. They're used here to
//------------------------------------------------------------------------------
const char* cmds[] = {"Heat", "Off", "Cool", "SP"};
const char box(255);  // filled box
const char deg(223);  // degree
const char open(220);  // open bottom box
const char hole(219); // empty box
const char omeg(244); // omega

// Initialize Variables
//------------------------------------------------------------------------------
bool debug     = false;
char mode      = 'O';
int coolPoint  = 80;
int heatPoint  = 65;
int setPoint   = 70;
int relayState = LOW;
int packetSize = 0;
float t        = 0;
float w        = 0;
float dt0      = 0;
float dt1      = 0;
int mot        = 8;
char packetBuffer[UDP_TX_PACKET_MAX_SIZE];       // buffer to hold incoming packet
char replyBuffer[] = "000\tO\t00\t0\t00\t000 ...";  // a string to send back
char l0buffer[16]  = "OFF  ||||       ";
char l1buffer[16]  = "  SP:70         ";
// global flag for Rotary knob
bool spMod         = false;



//------------------------------------------------------------------------------
// Ensure minimum off time
//------------------------------------------------------------------------------
void getMOT() {
   --mot;
   int  col = 8 - (mot % 4);
   l0buffer[col] = (mot < 4) ? omeg : hole;
   updateDisplay();
}
//------------------------------------------------------------------------------
TimedAction motThread    = TimedAction(7500, getMOT);

//------------------------------------------------------------------------------
// Store Status of temperature and relays
//------------------------------------------------------------------------------
void getStatus() {
   String rec  = String(round(t * 10)) + "\t"; // temperature to 1/10 of degree
   rec += mode; rec += "\t";    // thermostat mode
   rec += setPoint; rec += "\t"; // thermostat set point
   rec += relayState; rec += "\t";
   rec += mot; rec += "\t";
   rec += round(w * 10);
   rec.toCharArray(replyBuffer, rec.length() + 1);
}
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
void updateDisplay() {
   static char lcd0[] = "................";
   static char lcd1[] = "................";
   for (int i=0; i<16; i++) {
      if (lcd0[i] == l0buffer[i]) continue;
      lcd0[i] = l0buffer[i];
      lcd.setCursor(i,0);
      lcd.print(lcd0[i]);
   }
   for (int i=0; i<16; i++) {
      if (lcd1[i] == l1buffer[i]) continue;
      lcd1[i] = l1buffer[i];
      lcd.setCursor(i,1);
      lcd.print(lcd1[i]);
   }
}
//------------------------------------------------------------------------------

// measure the temperatrues and active/deactive relays as needed
//------------------------------------------------------------------------------
void control_environment() {
   sensors.requestTemperatures();
   w  = sensors.getTempFByIndex(0) + dt0; // read Room Temperature as fahrenheit
   t  = sensors.getTempFByIndex(1) + dt1; // read Flue Temperature as fahrenheit
   if (w < 32) w = 32;
   if (t < 32) t = 32;
   char result[6];  // 999.9
   dtostrf(w, 5,1, result);
   for(int i=0; i<5; i++)
      l0buffer[i+10] = result[i];

   dtostrf(t, 5,1, result);
   for(int i=0; i<5; i++)
      l1buffer[i+10] = result[i];

   updateRelays();
   updateDisplay();
}
//------------------------------------------------------------------------------
void updateRelays() {
   int tmp = LOW;  // default is system is off
   if (mot > 0) {
      tmp = relayState;

   } else if (mode == 'H') {
      if (relayState == LOW and t < (setPoint - 0.5))
         tmp = HIGH;
      else if (relayState == HIGH and t < (setPoint + 0.6))
         tmp = HIGH;

   } else if (mode == 'C') {
      if (relayState == LOW and t > (setPoint + 0.5))
         tmp = HIGH;
      else if (relayState == HIGH and t > (setPoint - 0.4))
         tmp = HIGH;
   }

   if (tmp != relayState) {
      relayState = tmp;
      digitalWrite(SSL_RELAY, relayState);
      if (relayState == LOW) {
         mot = 8;
         motThread.reset();
         for(int i=5; i<9; i++)
            l0buffer[i] = box;
      }
   }

   /*Serial.print(" Temp: ");
   Serial.print(t, 1);    //print the temperature
   Serial.print(" Wall: ");
   Serial.print(w, 1);    //print the temperature
   Serial.print(" Mode:");
   Serial.print(mode);
   Serial.print(" CoolPoint:");
   Serial.print(coolPoint);
   Serial.print(" HeatPoint:");
   Serial.print(heatPoint);
   Serial.print(" Set Point:");
   Serial.print(setPoint);
   Serial.print("  Relay:");
   Serial.print(relayState);
   Serial.print("  MOT:");
   Serial.print(mot);
   Serial.println("");*/
}
//------------------------------------------------------------------------------


/* Set actions to perform on separate timed intervals
 *------------------------------------------------------------------------------
 * http://playground.arduino.cc/Code/TimedAction
 * http://wiring.uniandes.edu.co/source/trunk/wiring/firmware/libraries/TimedAction
 * multi-task ie create a timers that will fire repeatedly every x ms */
TimedAction environment  = TimedAction(14947, control_environment);
//-------------------------------------------------------------------------------


//--------------------------------------------------------------------------------------------
void setup() {
   // initialize the relay pins as an output:
   pinMode(SSL_RELAY, OUTPUT);
   digitalWrite(SSL_RELAY, LOW);

   // start the Ethernet and UDP:
   Ethernet.begin(mac, ip);
   Udp.begin(localPort);

   // initialise 1602 LCD Display Module 16X2
   l0buffer[15] = deg;
   l1buffer[15] = deg;
   l1buffer[7]  = deg;
   for(int i=5; i<9; i++)
      l0buffer[i] = box;
   lcd.begin();
   lcd.backlight();
   lcd.clear();
   updateDisplay();

   if (debug) {
      Serial.begin(115200);
      Serial.println("Setup Complete");
   }
}
//--------------------------------------------------------------------------------------------

//--------------------------------------------------------------------------------------------
void loop() {
   readRotary();
   if (mot) motThread.check();
   if (!spMod) environment.check(); // dont react to user changing the setPoint

   // if there's data available, read a packet
   packetSize = Udp.parsePacket();
   if (packetSize) {
      Udp.read(packetBuffer, UDP_TX_PACKET_MAX_SIZE); // read the packet into packetBuffer

        /*   Serial.print("Received packet of size ");
           Serial.print(packetSize);
           Serial.print(" From ");
           IPAddress remote = Udp.remoteIP();
           for (int i = 0; i < 4; i++) {
              Serial.print(remote[i], DEC);
              if (i < 3) Serial.print(".");
           }
           Serial.print(", port ");
           Serial.print(Udp.remotePort());
           Serial.print(" Contents: ");
           Serial.println(packetBuffer); */

      if (packetBuffer[0] == 'r')
         getStatus();
      else {
         for (int i=0; i < 12; i++)
            obeyCommmad(cmds[i]);
      }

      // send a reply to the IP address and port that sent us the packet we received
      Udp.beginPacket(Udp.remoteIP(), Udp.remotePort());
      Udp.write(replyBuffer);
      Udp.endPacket();
      updateDisplay();
   }
}
//--------------------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Process incoming command
//------------------------------------------------------------------------------
void obeyCommmad(const char* cmd) {
   char *ret;
   ret = strstr(packetBuffer, cmd);
   if (ret == NULL) return;
   /*if (debug) {
      Serial.print("set command: ");
      Serial.println(cmd);
   }*/

   if (cmd == "SP") {
      setPoint = atoi( &packetBuffer[2] );
      newSetPoint();

   } else if (cmd == "Heat") {
      mode = 'H';
      setPoint = heatPoint;
      newSetPoint();
      for(int i=0; i<4; i++)
         l0buffer[i] = cmd[i];

   } else if (cmd == "Cool") {
      mode      = 'C';
      setPoint  = coolPoint;
      newSetPoint();
      for(int i=0; i<4; i++)
         l0buffer[i] = cmd[i];

   } else if (cmd == "Off") {
      mode      = 'O';
      for(int i=0; i<3; i++)
         l0buffer[i] = cmd[i];
      l0buffer[3] = ' ';
   }

   updateRelays();
   getStatus();
}
//------------------------------------------------------------------------------

// update new setPoint and display
//------------------------------------------------------------------------------
void newSetPoint() {
   if (setPoint > 95) setPoint = 95;
   if (setPoint < 55) setPoint = 55;
   if (mode == 'H' and heatPoint != setPoint)
      heatPoint = setPoint;
   if (mode == 'C' and coolPoint != setPoint)
      coolPoint = setPoint;

   char result[3];  // 999
   itoa(setPoint, result,10);
   l1buffer[5] = result[0];
   l1buffer[6] = result[1];
}
//------------------------------------------------------------------------------

// read rotary nob
//------------------------------------------------------------------------------
void readRotary() {
   static unsigned long spTimer  = 0;
   static int spCalc             = 70;
   static int oldPosition        = 0;
   static int spCount            = 0;

   int newPosition = spRotary.read();
   if (newPosition != oldPosition) {
      oldPosition = newPosition;
      spCount    += oldPosition < 0 ? -1 : 1;
      if (!spMod and abs(spCount) >= SP_STEP) {
         spMod  = true;
         spCalc = setPoint;
      }
      if (debug) {
      Serial.print("Mod: ");
      Serial.print(spMod);
      Serial.print(" new Pos: ");
      Serial.print(spCount);
      Serial.print(" spCalc: ");
      Serial.print(spCalc);
      Serial.print(" rotary: ");
      Serial.print((spCount / SP_STEP) + spCalc);
      Serial.print("  timer: ");
      Serial.println(millis() - spTimer);
      }
      if (spMod) setPoint = (spCount / SP_STEP) + spCalc;
      newSetPoint();
      updateDisplay();
      spTimer  = millis();
   } else if (spTimer and millis() - spTimer > 3000) {
      spTimer = 0;
      spCount = 0;
      spRotary.write(0);
      oldPosition = 0;
      if (spMod) {
         if (debug) {
            Serial.print("   New SetPoint: ");
            Serial.println(setPoint);
            Serial.println("~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~");
		     }
		     spMod  = false;
         spCalc = setPoint;
         updateRelays();
      }
      else if (debug) {
        Serial.println("   Rotary moved with being touched!");
        Serial.println("~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~");
      }
   }
}
//------------------------------------------------------------------------------
